/* === COPYRIGHT 2000 (C) ========================= TOM HENDRICK ===

   FILE:                                                     ether.c

   DESCRIPTION:
      Provides the functionality of operations on etherinfo_t structures
      including printing them in human and machine-readable formats.

   $REVISION:$
   
   LICENSE:
      This file, and the files included within the tipxd project are
      distributed under the GNU Public License(GPL).  Please read the
      LICENSE file included with this distribution or consult the license
      on the web at http://www.opensource.org/licenses/gpl-license.html

   =========== T I P X D . S O U R C E F O R G E . N E T =========== */

/* -----------------------------------------------------------------
   PREPROCESSOR INCLUDES
   ----------------------------------------------------------------- */
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <syslog.h>
#include <net/if.h>
#include <linux/if_packet.h>

#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "ether.h"
#include "main.h"
#include "log.h"

/* -----------------------------------------------------------------
   GLOBAL OR OTHERWISE EXPORTED DATA
   ----------------------------------------------------------------- */


/*=================================================================
FUNCTION: ether_address_to_str

DESCRIPTION:
   Prints an ethernet address to a static string buffer and returns a
   pointer to the buffer.

ARGUMENTS:
   char * - the array containing the hw address

RETURN TYPE:
   char * - a string version of the hw_address 
=================================================================*/
char *ether_address_to_str( const unsigned char *addr )
{
   static char buf[19];

   /* This is a little sprintf */
   snprintf( buf, 18, "%.2X:%.2X:%.2X:%.2X:%.2X:%.2X",
               addr[0], addr[1], addr[2],
               addr[3], addr[4], addr[5] );
   /* Return the string */
   return buf;
}

/* -----------------------------------------------------------------

   FUNCTION:                                          dump_ether( )

   DESCRIPTION:
      Dumps the contents of the etherinfo structure to the given File
      descriptor in a human-readable format.

   ARGUMENTS:
      FILE *fp - the file to dump to the ether information to.
      etherinfo_t *info - the info structure to dump to

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void dump_ether( FILE *fp, etherinfo_t *info )
{
   int i;
   if( !info ) return;
   // Dump the contents of info
   fprintf( fp, "NAME: %s\n", info->name );
   fprintf( fp, " INDEX  : %d\n", info->ifindex );
   fprintf( fp, " HWADDR : %s\n", ether_address_to_str(info->hwaddr));
   fprintf( fp, " PROTOS : (%d) %s%s%s%s\n",
            info->protocount,
            (info->protomask & PROTO_802_3_MASK)?"802.3 ":" ",
            (info->protomask & PROTO_802_2_MASK)?"802.2 ":" ",
            (info->protomask & PROTO_II_MASK)?"II ":" ",
            (info->protomask & PROTO_SNAP_MASK)?"snap":" " );
   for( i = 0; i < info->protocount; i++ )
      dump_proto( fp, &info->protocols[i] );
}

/* -----------------------------------------------------------------

   FUNCTION:                                          dump_proto( )

   DESCRIPTION:
      Dumps the contents of the protoinfo structure to the given File
      descriptor in a human-readable format.

   ARGUMENTS:
      FILE *fp - the file to dump to the proto information to.
      protoinfo_t *info - the info structure to dump to

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void dump_proto( FILE *fp, protoinfo_t *info )
{
   fprintf( fp, "  PROTO  : %s\n", info->name );
   if( info->selectinfo != NULL )
   {
   fprintf( fp, "  SOCKFD : %sOPEN", (info->selectinfo->fd< 0)?"NOT ":"" );
   }
   fprintf( fp, "  BYTES  : %ld\n", info->bytes_rcvd );
}

/* -----------------------------------------------------------------

   FUNCTION:                                          send_proto( )

   DESCRIPTION:
      Dumps the contents of the protoinfo structure to the given File
      descriptor in a machine-friendly format.

   ARGUMENTS:
      FILE *fp - the file to dump to the proto information to.
      protoinfo_t *info - the info structure to dump to

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void send_proto( int fd, protoinfo_t *info )
{
   size_t len;
   protoinfo_t p;
   char *ptr = (char *)&p;
   int retval = 0;
   int bytes;

   // Copy the information to a temporary variable
   len = sizeof( protoinfo_t );
   memcpy( &p, info, len );

   // Do some byte-order operating and blanking
   p.protocol   = htons( p.protocol );
   p.selectinfo = NULL;
   p.bytes_rcvd = htonl( p.bytes_rcvd );
   
   /* Okay, now do this all in one big write.  This piece of code is
    * annoying becuse it attempts to deal with the case of the EINTR value
    * of errno, which corresponds to an interrupted system call, where
    * there was no REAL error, but not all the bytes were written to the
    * file descriptor.
    */
   bytes = 0;
   do
   {
      retval = write( fd, ptr + bytes, len - bytes );
      if( retval < 0 ) { return; }
      bytes += retval;
   } while( bytes < len );
}

/* -----------------------------------------------------------------

   FUNCTION:                                          send_ether( )

   DESCRIPTION:
      Dumps the contents of the etherinfo structure to the given File
      descriptor in a machine-friendly format.

   ARGUMENTS:
      FILE *fp - the file to dump to the ether information to.
      etherinfo_t *info - the info structure to dump to

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void send_ether( int fd, etherinfo_t *info )
{
   size_t len;
   etherinfo_t e;
   char *ptr = (char *)&e;
   int retval = 0;
   int bytes;
   int i;

   // Copy the information to a temporary variable
   len = sizeof( etherinfo_t );
   memcpy( &e, info, len );

   // Do some byte-order operating and blanking
   e.ifindex    = htonl( e.ifindex );
   e.protocount = htonl( e.protocount );
   e.protomask  = htonl( e.protomask );
   e.next       = NULL;
   
   /* Okay, now do this all in one big write.  This piece of code is
    * annoying becuse it attempts to deal with the case of the EINTR value
    * of errno, which corresponds to an interrupted system call, where
    * there was no REAL error, but not all the bytes were written to the
    * file descriptor.
    */
   bytes = 0;
   do
   {
      retval = write( fd, ptr + bytes, len - bytes );
      if( retval < 0 ) { return; }
      bytes += retval;
   } while( bytes < len );


   /* dump the protocol information to the FD as well */
   for( i = 0; i < info->protocount; i++ )
         send_proto( fd, &info->protocols[i] );

}


/* -----------------------------------------------------------------

   FUNCTION:                                           add_ether( )

   DESCRIPTION:
      Adds an Ethernet device structure to the system list

   ARGUMENTS:
      etherinfo_t *info - the info structure to add to the sysinfo

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void add_ether( etherinfo_t *info )
{
   etherinfo_t *ptr;
   if( NULL == info ) return;

   // Okay, now, see if we are adding to the end of a list, or creating it
   if( sysinfo.etherinfo == NULL )
   {
      sysinfo.etherinfo = info;
   }
   else
   {
      for( ptr = sysinfo.etherinfo; ptr->next != NULL; ptr = ptr->next );
      ptr->next = info;
      info->next = NULL;
   }
}

/* -----------------------------------------------------------------

   FUNCTION:                                         verify_ether( )

   DESCRIPTION:
      Adds an Ethernet device structure to the system list

   ARGUMENTS:
      etherinfo_t *info - the info structure to add to the sysinfo

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
int verify_ether( etherinfo_t *info )
{
   int fd;
   struct ifreq req;

   if( NULL == info )
   {
      return -1;
   }

   /* Ensure that the name is not blank */
   if( info->name == '\0' )
   {
      fprintf( stderr, "Name not set for device!\n" );
      return -1;
   }

   /* Ensure that the protocol count is > 0 */
   if( info->protocount == 0 )
   {
      fprintf( stderr, "No protocols found for device!\n" );
      return -1;
   }

   /* Attempt to get the hardware address for the device */
   fd = socket( PF_INET, SOCK_DGRAM, 0 );
   if( fd < 0 )
   {
      /* Ooops, no such device */
      perror( "socket" );
      return -1;
   }
   
   /* Get the ifreq structure */
   strncpy( req.ifr_name, info->name, 16 );
   if( ioctl( fd, SIOCGIFHWADDR, &req ) < 0 )
   {
      fprintf( stderr, "ioctl( ) failed for device : %s\n",
               strerror( errno ) );
      return -1;
   }

   memcpy( info->hwaddr, req.ifr_hwaddr.sa_data, ETHER_ADDR_LEN );

   close( fd );
   
   return 0; 
}

/* -----------------------------------------------------------------

   FUNCTION:                                             open_ipx( )

   DESCRIPTION:
      Opens all protocol sockets for this ethernet device and binds them to
      the physical interface.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to a etherinfo_t * structure

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void open_ipx( struct _selectinfo_t *info, void *arg )
{
   etherinfo_t *ether;
   protoinfo_t *proto;
   struct sockaddr_ll addr;
   struct packet_mreq mreq;
   struct ifreq ireq;
   int fd;
   int retval;
   int index;

   /* Typecast the arg */
   index = *((int *)arg);
   ether = (etherinfo_t *)info->data;
   proto = (protoinfo_t *)&(ether->protocols[index]);

   /* Attempt to open the socket */
   fd = socket( PF_PACKET, SOCK_RAW, htons( proto->protocol ) );
   if( fd < 0 )
   {
      /* Error opening the socket */
      tipxd_log( LOG_ERR, "Unable to open IPX socket : %m" );
      return;
   }

   /* Attempt to get the device index for this device */
   memcpy( ireq.ifr_name, ether->name, IFNAMSIZ );
   retval = ioctl( fd, SIOCGIFINDEX, &ireq, sizeof( struct ifreq ) );

   if( retval < 0 )
   {
      tipxd_log( LOG_ERR, "Unable to get interface index : %m" );
      close( fd );
      return;
   }

   ether->ifindex = ireq.ifr_ifindex;

   /* Okay, attempt to bind the socket */
   addr.sll_family   = AF_PACKET;
   addr.sll_ifindex  = ether->ifindex;
   addr.sll_protocol = htons( proto->protocol );
   addr.sll_hatype   = ARPHRD_ETHER;
   if( proto->flags & PROTO_BROADCAST )
   {
      addr.sll_pkttype = PACKET_BROADCAST;
   }
   else
   {
      addr.sll_pkttype = PACKET_OTHERHOST;
   }
   addr.sll_halen = ETHER_ADDR_LEN;
   memcpy( addr.sll_addr, ether->hwaddr, ETHER_ADDR_LEN );

   /* Attempt to bind the socket to the interface */
   retval = bind( fd, (struct sockaddr *)&addr, sizeof(addr) );
   if( retval < 0 )
   {
      /* Problem */
      tipxd_log( LOG_ERR, "Unable to bind socket to device %s : %m",
               ether->name );
      close( fd );
      return;
   }

   /* Socket opened and bound, set the flags for promiscuous mode */
   mreq.mr_ifindex = ether->ifindex;
   mreq.mr_type = PACKET_MR_PROMISC;
   mreq.mr_alen = ETHER_ADDR_LEN;
   memcpy( mreq.mr_address, ether->hwaddr, ETHER_ADDR_LEN );

   retval = setsockopt( fd, SOL_PACKET, PACKET_ADD_MEMBERSHIP,
                  &mreq, sizeof( mreq ) );
   if( retval < 0 )
   {
      tipxd_log( LOG_ERR, "Unable to set promiscuous mode : %m" );
      close( fd );
      return;
   }

   /* We are open, listening, and bound, set flags and return */
   tipxd_log( LOG_INFO, "IPX socket %s (%s) opened on fd %d",
              ether->name, proto->name, fd );

   info->fd = fd;
}


/* -----------------------------------------------------------------

   FUNCTION:                                            write_ipx( )

   DESCRIPTION:
      Writes the given data buffer to the IPX socket.  Should an error
      occur, calls the error handler.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to an integer argument, representing the index of
         the protoinfo in the etherinfo array.
      char * - the data buffer to write.
      size_t - the size in bytes of the buffer.

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void write_ipx( struct _selectinfo_t *info, void *arg, char *buf, size_t s )
{
   etherinfo_t *ether;
   protoinfo_t *proto;
   int index;
   int retval;
   size_t bytes;

   index = *((int *)arg);
   ether = (etherinfo_t *)info->data;
   proto = (protoinfo_t *)(&ether->protocols[index]);

   printf( "Writing %d bytes to IPX socket %s\n",
            s, ether->name );

   /* Okay, now do this all in one big write.  This piece of code is
    * annoying becuse it attempts to deal with the case of the EINTR value
    * of errno, which corresponds to an interrupted system call, where
    * there was no REAL error, but not all the bytes were written to the
    * file descriptor.
    */
   bytes = 0;
   do
   {
      retval = write( info->fd, buf + bytes, s - bytes );
      if( retval < 0 ) { info->error_op( info, arg, errno ); return; }
      bytes += retval;
   } while( bytes < s );
}

/* -----------------------------------------------------------------

   FUNCTION:                                             read_ipx( )

   DESCRIPTION:
      Reads a packet from the IPX interface given, and attempts to route it
      using the route function.

   ARGUMENTS:
      selectinfo_t * - the selectinfo structure.
      void * - typecast to a etherinfo_t * structure

   RETURN VALUE:
      N/A

   ----------------------------------------------------------------- */
void read_ipx( struct _selectinfo_t *info, void *arg )
{
   char *buf;
   size_t len;
   int i;
   etherinfo_t *ether;
   struct ethhdr *header;
   peerinfo_t *src_peer, *dst_peer;
   
   ether = (etherinfo_t *)info->data;
   i     = *(int *)arg;

   len   = sizeof( datapkt_t ) + ETH_FRAME_LEN;
   buf   = (char *)malloc( len );
   memset( buf, '\0', len );

   /* Attempt to read an entire buffer */
   len = recvfrom( info->fd, buf + sizeof(datapkt_t), len, 0, NULL, NULL );

   header = (struct ethhdr *)(buf + sizeof(datapkt_t));
   src_peer = find_peer( header->h_source );
   dst_peer = find_peer( header->h_dest );

   tipxd_log( LOG_INFO, "Read %d bytes from IPX: %s (%s) %s",
               len, ether->name, ether->protocols[i].name,
               ether_address_to_str( header->h_source ));

   tipxd_log( LOG_INFO, "   Dest is %s",
               ether_address_to_str( header->h_dest ) );
   /* If the src_peer is not registered, then we nede to register it */
   if( src_peer == NULL )
   {
      src_peer = (peerinfo_t *)malloc( sizeof( peerinfo_t ) );
      memset( src_peer, '\0', sizeof( peerinfo_t ) );

      memcpy( src_peer->address, header->h_source, ETHER_ADDR_LEN );
   }

   if( src_peer->location == PEER_LOC_UNREGISTERED )
   {
      src_peer->location = PEER_LOC_LOCAL;
      src_peer->destination.ether = ether;
   }

   /* Attempt to route the packet */
   route( src_peer, dst_peer, NULL, buf, len + sizeof( datapkt_t ));

   free( buf );
}
